;;; spam-gmane.el --- functions for administrating spam on Gmane
;; Copyright (C) 2004 Free Software Foundation, Inc.

;; Author: Lars Magne Ingebrigtsen <larsi@gnus.org>
;; Keywords: mail news

;; This file is not part of GNU Emacs.

;; gmane.el is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; gmane.el is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with gmane.el; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;; To use this, you have to use the latest CVS version of Gnus.
;;
;; Put something like the following in your .emacs file:
;;
;; (require 'cl)
;; (push "~/gmadmin" load-path)
;; (push 'spam-gmane-set-nov gnus-summary-mode-hook)
;; (autoload 'spam-gmane-set-nov "spam-gmane")
;; (setq spam-gmane-user "something"
;;       spam-gmane-password "password")
;;
;; Put the following (`G p') in the group parameters for
;; gmane.spam.reported.spam:
;;
;; ((gnus-summary-line-format "%U%uS: %uT %-40,40s %[%-25,25uG %-20,20uA%z%I%(%-23,23f%]%)\n"))
;;
;; This will give you an overview with the data required.  It's very
;; wide, so you need to expand your Emacs to a width of about 120.
;;
;; I usually select the group when there have been a few hundred
;; reports done.  You'll find that most of the reports (valid or
;; invalid) are done from a handful of hosts, so sorting the messages
;; by submitter address (`C-c C-s C-p') is good for doing a first
;; pass, or `C-c C-s C-r' to sort by the number of submissions.  When
;; it gets time to get down to reports that require more careful
;; handling, I find that sorting by group (`C-c C-s C-g') is most
;; helpful.
;;
;; The commands are basically `a' for "approve thread", `r' for
;; "reject thread", and `M-x spam-gmane-submit-reports' to submit the
;; reports.  There's also `M-x spam-gmane-submit-reports-async' that
;; does the same as the other command, but submits reports in the
;; background.
;;
;; In addition, there's the following commands:
;; `s': approve and select next article
;; `t': reject and select next article
;; (The idea here being that if you need to read through articles, you
;; have one finger on the `s' and one finger on the `t'.)
;;
;; `b' will deny the IP address of your choice from submitting
;; reports.  Only to be used for addresses that have misreported a
;; *lot*.

;;; Code:

(require 'gnus-sum)
(require 'cl)
(require 'gnus-art)
(require 'mm-url)
(require 'dns)

(defvar spam-gmane-user "")
(defvar spam-gmane-password "")

(defun spam-gmane-submit-reports ()
  "Submit all reports to Gmane."
  (interactive)
  (dolist (article spam-gmane-rejected)
    (when (gnus-summary-article-header article)
      (let ((id
	     (cdr (assq 'Id (mail-header-extra
			     (gnus-summary-article-header article))))))
	(spam-gmane-ping "reject" "spam"
			 (car id)
			 (cadr id)
			 article)
	(ignore-errors
	  (gnus-summary-mark-article article gnus-canceled-mark))
	(spam-gmane-remove-mark-1 article))))
  (dolist (article spam-gmane-approved)
    (when (gnus-summary-article-header article)
      (let ((id
	     (cdr (assq 'Id (mail-header-extra
			     (gnus-summary-article-header article))))))
	(spam-gmane-ping "accept" "spam"
			 (car id)
			 (cadr id)
			 article)
	(ignore-errors
	  (gnus-summary-mark-article article gnus-canceled-mark))
	(spam-gmane-remove-mark-1 article)))))

;; (add-hook 'gnus-article-prepare-hook 'spam-gmane-massage-article)

(setq gnus-alter-header-function 'spam-gmane-massage-header)
(add-hook 'gnus-summary-mode-hook 'spam-gmane-set-nov)

(defun spam-gmane-setup-summary ()
  (use-local-map (copy-keymap gnus-summary-mode-map))
  (local-set-key "\C-c\C-s\C-g" 'spam-gmane-sort-by-group)
  (local-set-key "\C-c\C-s\C-r" 'spam-gmane-sort-by-reports)
  (local-set-key "\C-c\C-s\C-p" 'spam-gmane-sort-by-reporter)
  (local-set-key "/r" 'spam-gmane-limit-to-reporter)
  (local-set-key "/z" 'spam-gmane-limit-to-zero-reports)
  (local-set-key "u" 'spam-gmane-remove-mark)
  (local-set-key "A" 'spam-gmane-approve)
  (local-set-key "R" 'spam-gmane-reject)
  (local-set-key "a" 'spam-gmane-approve-thread)
  (local-set-key "b" 'spam-gmane-ban-reporter)
  (local-set-key "r" 'spam-gmane-reject-thread)
  (local-set-key "s" 'spam-gmane-approve-and-select)
  (local-set-key "t" 'spam-gmane-reject-and-select))

(defun spam-gmane-limit-to-reporter ()
  "Limit to articles that the current reporter has reported."
  (interactive)
  (let ((articles nil)
	(queue (cdr (assq 'Queue (mail-header-extra
				  (gnus-summary-article-header)))))
	reporter)
    (unless (= (length queue) 1)
      (error "Invalid reporter length"))
    (setq reporter (nth 2 (car queue)))
    (dolist (info gnus-newsgroup-data)
      (setq info (nth 3 info))
      (when (vectorp info)
	(dolist (report (cdr (assq 'Queue (mail-header-extra info))))
	  (when (equalp reporter (nth 2 report))
	    (push (mail-header-number info) articles)))))
    (gnus-summary-limit articles)))

(defun spam-gmane-limit-to-zero-reports ()
  "Limit to articles that have zero reports."
  (interactive)
  (let ((articles nil))
    (dolist (info gnus-newsgroup-data)
      (setq info (nth 3 info))
      (when (vectorp info)
	(when (zerop (length (cdr (assq 'Queue (mail-header-extra info)))))
	  (push (mail-header-number info) articles))))
    (gnus-summary-limit articles)))

(defun spam-gmane-approve-and-select ()
  "Approve and select next."
  (interactive)
  (spam-gmane-approve 1)
  (gnus-summary-select-article))

(defun spam-gmane-reject-and-select ()
  "Reject and select next."
  (interactive)
  (spam-gmane-reject 1)
  (gnus-summary-select-article))

(defun spam-gmane-approve (n &optional unmark remove)
  "Set the process mark on the next N articles.
If N is negative, mark backward instead.  If UNMARK is non-nil, remove
the process mark instead.  The difference between N and the actual
number of articles marked is returned."
  (interactive "P")
  (setq n (prefix-numeric-value n))
  (let ((backward (< n 0))
	(n (abs n)))
    (while (and
	    (> n 0)
	    (cond
	     (unmark
	      (spam-gmane-reject-mark
	       (gnus-summary-article-number)))
	     (remove
	      (spam-gmane-remove-mark-1
	       (gnus-summary-article-number)))
	     (t
	      (spam-gmane-approve-mark (gnus-summary-article-number))))
	    (zerop (gnus-summary-next-subject (if backward -1 1) nil t)))
      (setq n (1- n)))
    (when (/= 0 n)
      (gnus-message 7 "No more articles"))
    n))

(defun spam-gmane-approve-thread ()
  "Marks all articles downwards in this thread."
  (interactive)
  (let ((score (assoc 'score gnus-summary-mark-positions)))
    (when score
      (setcdr score 2)))
  (save-excursion
    (ignore-errors
      (gnus-summary-mark-as-read-forward 1)))
  (let ((level (gnus-summary-thread-level)))
    (while (and (spam-gmane-approve-mark
		 (gnus-summary-article-number))
		(zerop (gnus-summary-next-subject 1 nil t))
		(> (gnus-summary-thread-level) level)))))

(defun spam-gmane-reject-thread ()
  "Reject all articles downwards in this thread."
  (interactive)
  (save-excursion
    (gnus-summary-mark-as-read-forward 1))
  (let ((level (gnus-summary-thread-level)))
    (while (and (spam-gmane-reject-mark
		 (gnus-summary-article-number))
		(zerop (gnus-summary-next-subject 1 nil t))
		(> (gnus-summary-thread-level) level)))))

(defun spam-gmane-reject (n)
  (interactive "P")
  (spam-gmane-approve n t))

(defun spam-gmane-remove-mark (n)
  (interactive "P")
  (spam-gmane-approve n nil t))

(defun spam-gmane-approve-mark (article)
  "Set the process mark on ARTICLE and update the summary line."
  (setq spam-gmane-rejected (delq article spam-gmane-rejected))
  (setq spam-gmane-approved
	(cons article
	      (delq article spam-gmane-approved)))
  (when (gnus-summary-goto-subject article)
    (gnus-summary-show-thread)
    (gnus-summary-goto-subject article)
    (spam-gmane-update-mark article)))

(defun spam-gmane-reject-mark (article)
  "Reject the process mark on ARTICLE and update the summary line."
  (setq spam-gmane-approved (delq article spam-gmane-approved))
  (setq spam-gmane-rejected
	(cons article
	      (delq article spam-gmane-rejected)))
  (when (gnus-summary-goto-subject article)
    (gnus-summary-show-thread)
    (gnus-summary-goto-subject article)
    (spam-gmane-update-mark article)))

(defun spam-gmane-remove-mark-1 (article)
  "Reject the process mark on ARTICLE and update the summary line."
  (setq spam-gmane-approved (delq article spam-gmane-approved))
  (setq spam-gmane-rejected (delq article spam-gmane-rejected))
  (when (gnus-summary-goto-subject article)
    (gnus-summary-show-thread)
    (gnus-summary-goto-subject article)
    (spam-gmane-update-mark article)))

(defun spam-gmane-update-mark (article)
  "Update the secondary (read, process, cache) mark."
  (let ((buffer-read-only nil))
    (beginning-of-line)
    (subst-char-in-region
     (point) (1+ (point)) (following-char)
     (cond
      ((memq article spam-gmane-approved)
       ?$)
      ((memq article spam-gmane-rejected)
       ?*)
      (t
       ? ))))
  t)

(defun spam-gmane-set-nov ()
  (setq nntp-nov-is-evil (spam-gmane-group-p))
  (when (spam-gmane-group-p)
    (spam-gmane-setup-summary)
    (setq spam-gmane-approved nil
	  spam-gmane-rejected nil)))

(defun spam-gmane-massage-header (header)
  (when (string-match "gmane.spam.reported" gnus-newsgroup-name)
    (save-excursion
      (let ((queue nil)
	    (score 0)
	    line)
	(goto-char (point-min))
	(while (search-forward "\nX-Gmane-Queue:" nil t)
	  (setq line (split-string (buffer-substring (point) (line-end-position))))
	  (message "Resolving %s" (nth 4 line))
	  (push (list (nth 0 line) (nth 2 line)
		      (or (dns-query-cached (nth 4 line) nil nil t)
			  (nth 4 line))
		      (nth 3 line))
		queue))
	(goto-char (point-max))
	(when (search-backward "\nX-Spam-Report:" nil t)
	  (goto-char (match-end 0))
	  (setq score (read (current-buffer)))
	  (unless (numberp score)
	    (setq score 0)))
	(mail-header-set-extra
	 header (list (list 'Id (nth 1 (car queue)) (nth 3 (car queue)))
		      (list 'Ip (nth 4 line))
		      (list 'Score score)
		      (cons 'Queue (spam-gmane-collate-reports queue))))))))

(defun spam-gmane-find-prev (address reports)
  (let (out)
    (dolist (report reports)
      (when (equalp address (nth 2 report))
	(setq out report)))
    out))

(defun spam-gmane-collate-reports (reports)
  (let ((out nil)
	prev)
    (dolist (report reports)
      (if (setq prev (spam-gmane-find-prev (nth 2 report) out))
	  (unless (string= (car prev) (car report))
	    (setq out (delq prev out)))
	(push report out)))
    (nreverse out)))

(defvar spam-gmane-approved nil)
(defvar spam-gmane-rejected nil)

(defun gnus-user-format-function-S (header)
  (cond
   ((memq (mail-header-number header) spam-gmane-approved)
    "$")
   ((memq (mail-header-number header) spam-gmane-rejected)
    "*")
   (t
    " ")))

(defun gnus-user-format-function-T (header)
  (let ((score (cadr (assq 'Score (mail-header-extra header)))))
    (format "%4.1f" (if (numberp score) score 0))))

(defvar spam-gmane-group-name-abbrevs
  '(("gmane" . "g") ("comp" . "c") ("linux" . "l") ("culture" . "cul")
    ("recreation" . "rec") ("science" . "sci")
    ("audio" . "au") ("graphics" . "gfx") 
    ("hardware" . "hw") ("internationalization" . "i18n"))
  "Maps group name components to abbreviations for use in %uG.")
   
(defun gnus-user-format-function-G (header)
  (let* ((group (cadr (cadr (assq 'Queue (mail-header-extra header)))))
	 (parts (split-string (or group "") "[.]")))
    (mapconcat 
     #'(lambda (part) 
	 (or (cdr (assoc part spam-gmane-group-name-abbrevs))
	     part))
     parts ".")))

(defun gnus-user-format-function-A (header)
  (let ((queue (cdr (assq 'Queue (mail-header-extra header)))))
    (if (= (length queue) 1)
	(format "%s" (nth 2 (car queue)))
      (format "%d reports" (length queue)))))

(defun spam-gmane-group-p ()
  (member (gnus-group-real-name gnus-newsgroup-name)
	  '("gmane.spam.reported.spam" "gmane.spam.reported.unspam"
	    "gmane.comp.gnome.apps.epiphany"
	    "gmane.comp.gcc.g++.bugs"
	    "gmane.comp.lang.smalltalk.gnu.bugs"
	    "gmane.editors.nano.general"
	    "gmane.linux.kernel.firewire.devel"
	    "gmane.emacs.easymacs.user"
	    "gmane.comp.search.nutch.devel"
	    "gmane.network.serveez.devel"
	    "gmane.org.fsf.europe.translators"
	    "gmane.emacs.bugs"
	    "gmane.comp.archivers.dump.devel"
	    "gmane.comp.handhelds.netbsd"
	    "gmane.comp.hardware.pciids.devel")))

(defun spam-gmane-massage-article ()
  (when (spam-gmane-group-p)
    (let ((values
	   (split-string (gnus-fetch-original-field "x-gmane-queue") " "))
	  (buffer-read-only nil))
      (goto-char (point-min))
      (insert (format "X-Report-Info: %s\nX-Reporter: %s\n"
		      (nth 2 values)
		      (or (dns-query-cached (nth 4 values) nil nil t)
			  (nth 4 values)))))))

(defun spam-gmane-ping (type command group article queued-article)
  (message "%s %s:%s..." type group article)
  (save-excursion
    (set-buffer (get-buffer-create "*gmane spam*"))
    (goto-char (point-max))
    (insert "\n\n")
    (let ((url-request-method nil)
          (url-request-data nil))
          (mm-url-insert-file-contents
           (format
            "http://gmane.org/handle.php?group=%s&command=%s&article=%s&type=%s&user=%s&password=%s&queued_article=%s"
            (mm-url-form-encode-xwfu group)
            command article type spam-gmane-user spam-gmane-password
            queued-article)))))

(defun spam-gmane-reports (h)
  (length (cdr (assq 'Queue (mail-header-extra h)))))
  
(defsubst gnus-article-sort-by-reports (h1 h2)
  "Sort articles by article number."
  (> (spam-gmane-reports h1)
     (spam-gmane-reports h2)))

(defun gnus-thread-sort-by-reports (h1 h2)
  "Sort threads by root article number."
  (gnus-article-sort-by-reports
   (gnus-thread-header h1) (gnus-thread-header h2)))

(defun spam-gmane-sort-by-reports (&optional reverse)
  "Sort the summary buffer by reports number."
  (interactive "P")
  (gnus-summary-sort 'reports reverse))


(defun spam-gmane-reporter (h)
  (or (nth 2 (cadr (assq 'Queue (mail-header-extra h)))) ""))
  
(defsubst gnus-article-sort-by-reporter (h1 h2)
  "Sort articles by article number."
  (string< (spam-gmane-reporter h1)
	   (spam-gmane-reporter h2)))

(defun gnus-thread-sort-by-reporter (h1 h2)
  "Sort threads by reporter."
  (gnus-article-sort-by-reporter
   (gnus-thread-header h1) (gnus-thread-header h2)))

(defun spam-gmane-sort-by-reporter (&optional reverse)
  "Sort the summary buffer by reporter."
  (interactive "P")
  (gnus-summary-sort 'reporter reverse))


(defun spam-gmane-group (h)
  (or (cadr (assq 'Id (mail-header-extra h))) ""))
  
(defun spam-gmane-ip (h)
  (or (cadr (assq 'Ip (mail-header-extra h))) ""))
  
(defsubst gnus-article-sort-by-group (h1 h2)
  "Sort articles by article number."
  (string< (spam-gmane-group h1)
	   (spam-gmane-group h2)))

(defun gnus-thread-sort-by-group (h1 h2)
  "Sort threads by reporter."
  (gnus-article-sort-by-group
   (gnus-thread-header h1) (gnus-thread-header h2)))

(defun spam-gmane-sort-by-group (&optional reverse)
  "Sort the summary buffer by group."
  (interactive "P")
  (gnus-summary-sort 'group reverse))

(defun spam-gmane-ban-reporter (ip)
  "Ban the IP address of the reporter."
  (interactive (list (spam-gmane-ip (gnus-summary-article-header))))
  (setq ip (read-from-minibuffer "Ban IP (prefix): " ip))
  (save-excursion
    (set-buffer (get-buffer-create "*gmane spam*"))
    (goto-char (point-max))
    (insert "\n\n")
    (mm-url-insert-file-contents
     (format
      "http://gmane.org/handle.php?command=ban&user=%s&password=%s&ip=%s"
      spam-gmane-user spam-gmane-password ip))))

(defun spam-gmane-submit-reports-async ()
  "Submit all reports to Gmane using an asynchronous curl process."
  (interactive)
  (loop with process-connection-type = nil
	with process = (start-process "curl" (get-buffer-create "*curl*")
				      "curl" "-K" "-" "-s" "-S")
	with fmt = 
	(concat "http://gmane.org/handle.php?group=%s&command=%s&"
		"article=%s&type=%s&user=%s&password=%s&queued_article=%s")
	for (list type) in `((,spam-gmane-rejected "reject")
			     (,spam-gmane-approved "accept"))
	do
	(dolist (article list)
	  (when (gnus-summary-article-header article)
	    (let* ((id
		    (cdr (assq 'Id (mail-header-extra
				    (gnus-summary-article-header article)))))
		   (url (format fmt
				(mm-url-form-encode-xwfu (car id))
				"spam" (cadr id) type 
				spam-gmane-user spam-gmane-password article)))
	      (process-send-string process (format "url = %s\n" url))
	      (ignore-errors
		(gnus-summary-mark-article article gnus-canceled-mark))
	      (spam-gmane-remove-mark-1 article))))
	finally (process-send-eof process)))

(provide 'spam-gmane)

;;; spam-gmane.el ends here
